/*
 * Copyright (c) 1999 by David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * The "dd" built-in command.
 */

#include "message.h"
#include "sash.h"


#define	PAR_NONE	0
#define	PAR_IF		1
#define	PAR_OF		2
#define	PAR_BS		3
#define	PAR_COUNT	4
#define	PAR_SEEK	5
#define	PAR_SKIP	6


typedef	struct
{
	const char *	name;
	int		value;
} PARAM;


static const PARAM	params[] =
{
	{"if",		PAR_IF},
	{"of",		PAR_OF},
	{"bs",		PAR_BS},
	{"count",	PAR_COUNT},
	{"seek",	PAR_SEEK},
	{"skip",	PAR_SKIP},
	{NULL,		PAR_NONE}
};


static	long	getNum(const char * cp);


void
do_dd(int argc, const char *const *argv)
{
	const char *	str;
	const PARAM *	par;
	const char *	inFile;
	const char *	outFile;
	char *		cp;
	int		inFd;
	int		outFd;
	int		inCc;
	int		outCc;
	int		blockSize;
	long		count;
	long		seekVal;
	long		skipVal;
	long		intotal;
	long		outTotal;
	char *		buf;
	char		localBuf[BUF_SIZE];

	inFile = NULL;
	outFile = NULL;
	seekVal = 0;
	skipVal = 0;
	blockSize = 512;
	count = 0x7fffffff;

	while (--argc > 0)
	{
		str = *++argv;
		cp = strchr(str, '=');

		if ( !cp )
		{
			errmsg( "dd: invalid argument" );
			return;
		}

		*cp++ = '\0';

		for (par = params; par->name; par++)
		{
			if (strcmp(str, par->name) == 0)
				break;
		}

		switch (par->value)
		{
			case PAR_IF:
				if (inFile)
				{
					errmsg( "dd: multiple input files illegal" );
					return;
				}
	
				inFile = cp;
				break;

			case PAR_OF:
				if (outFile)
				{
					errmsg( "dd: multiple output files illegal" );
					return;
				}

				outFile = cp;
				break;

			case PAR_BS:
				blockSize = getNum(cp);

				if (blockSize <= 0)
				{
					errmsg( "dd: bad block size value" );
					return;
				}

				break;

			case PAR_COUNT:
				count = getNum(cp);

				if (count < 0)
				{
					errmsg( "dd: bad count value" );
					return;
				}

				break;

			case PAR_SEEK:
				seekVal = getNum(cp);

				if (seekVal < 0)
				{
					errmsg( "dd: bad seek value" );
					return;
				}

				break;

			case PAR_SKIP:
				skipVal = getNum(cp);

				if (skipVal < 0)
				{
					errmsg( "dd: bad skip value" );
					return;
				}

				break;

			default:
				errmsg( "dd: unknown parameter" );
				return;
		}
	}

	if ( !inFile )
	{
		errmsg( "dd: no input file specified" );
		return;
	}

	if ( !outFile )
	{
		errmsg( "dd: no output file specified" );
		return;
	}

	buf = localBuf;

	if (blockSize > sizeof(localBuf))
	{
		buf = alloca(blockSize);

		if ( !buf )
		{
			errmsg( "dd: buffer allocation failure: %s", strerror(errno) );
			return;
		}
	}

	intotal = 0;
	outTotal = 0;

	inFd = open(inFile, 0);

	if (inFd < 0)
	{
		errmsg( "dd: error opening %s: %s", inFile, strerror(errno) );
		return;
	}

	outFd = creat(outFile, 0666);

	if (outFd < 0)
	{
		errmsg( "dd: error opening %s: %s", outFile, strerror(errno) );
		close(inFd);

		return;
	}

	if (skipVal)
	{
		if (lseek(inFd, skipVal * blockSize, 0) < 0)
		{
			while (skipVal-- > 0)
			{
				inCc = read(inFd, buf, blockSize);

				if (inCc < 0)
				{
					errmsg( "dd: error reading %s while skipping: %s", inFile, strerror(errno) );
					goto cleanup;
				}

				if (inCc == 0)
				{
					errmsg( "dd: %s: unexpected end of file while skipping", inFile );
					goto cleanup;
				}
			}
		}
	}

	if (seekVal)
	{
		if (lseek(outFd, seekVal * blockSize, 0) < 0)
		{
			errmsg( "dd: error seeking %s: %s", outFile, strerror(errno) );
			goto cleanup;
		}
	}

	while ((inCc = read(inFd, buf, blockSize)) > 0)
	{
		intotal += inCc;
		cp = buf;

		if (intFlag)
		{
			errmsg( "dd: interrupted" );
			goto cleanup;
		}

		while (inCc > 0)
		{
			outCc = write(outFd, cp, inCc);

			if (outCc < 0)
			{
				errmsg( "dd: error writing %s: %s", outFile, strerror(errno) );
				goto cleanup;
			}

			outTotal += outCc;
			cp += outCc;
			inCc -= outCc;
		}
	}

	if (inCc < 0)
		errmsg( "dd: error reading %s: %s", inFile, strerror(errno) );

cleanup:
	if ( close(inFd) < 0 )
		errmsg( "dd: error closing %s: %s", inFile, strerror(errno) );

	if (close(outFd) < 0)
		errmsg( "dd: error closing %s: %s", outFile, strerror(errno) );

	if ( intotal || outTotal )
	{
		printf("%ld+%d records in\n", intotal / blockSize,
			(intotal % blockSize) != 0);

		printf("%ld+%d records out\n", outTotal / blockSize,
			(outTotal % blockSize) != 0);
	}
}


/*
 * Read a number with a possible multiplier.
 * Returns -1 if the number format is illegal.
 */
static long
getNum(const char * cp)
{
	long	value;

	if (!isDecimal(*cp))
		return -1;

	value = 0;

	while (isDecimal(*cp))
		value = value * 10 + *cp++ - '0';

	switch (*cp++)
	{
		case 'k':
			value *= 1024;
			break;

		case 'b':
			value *= 512;
			break;

		case 'w':
			value *= 2;
			break;

		case '\0':
			return value;

		default:
			return -1;
	}

	if (*cp)
		return -1;

	return value;
}

/* END CODE */
